from logging import getLogger
import time
from clickzetta.client import Client
from clickzetta.bulkload.bulkload_enums import BulkLoadMetaData, BulkLoadCommitOptions, BulkLoadState, \
    BulkLoadCommitMode
from clickzetta.bulkload.bulkload_writer import BulkLoadWriter

_logger = getLogger(__name__)


class BulkLoadStream:
    def __init__(self, meta_data: BulkLoadMetaData, client: Client):
        self.meta_data = meta_data
        self.client = client

    def get_stream_id(self):
        return self.meta_data.get_stream_id()

    def get_operation(self):
        return self.meta_data.get_operation()

    def get_stream_state(self):
        return self.meta_data.get_state()

    def get_sql_error(self):
        return self.meta_data.get_sql_error_msg()

    def get_schema(self):
        return self.meta_data.get_schema_name()

    def get_table(self):
        return self.meta_data.get_table()

    def get_record_keys(self):
        return self.meta_data.get_record_keys()

    def get_partition_specs(self):
        return self.meta_data.get_partition_specs()

    def open_writer(self, partition_id: int):
        config = self.client.open_bulkload_stream_writer(self.meta_data.get_instance_id(),
                                                         self.meta_data.get_workspace(),
                                                         self.meta_data.get_schema_name(),
                                                         self.meta_data.get_table_name(),
                                                         self.meta_data.get_stream_id(), partition_id)

        return BulkLoadWriter(self.client, self.meta_data, config, partition_id)

    def commit(self, options: BulkLoadCommitOptions):
        _logger.info("Committing BulkLoadStream:" + self.meta_data.get_stream_id())
        self.client.commit_bulkload_stream(self.meta_data.get_instance_id(), self.meta_data.get_workspace(),
                                           self.meta_data.get_schema_name(), self.meta_data.get_table_name(),
                                           self.meta_data.get_stream_id(), options.workspace, options.vc,
                                           BulkLoadCommitMode.COMMIT_STREAM)

        state = BulkLoadState.COMMIT_SUBMITTED
        sql_error_msg = ''
        # commit waiting max time is 5 hours.so try to get result every 10s for 1800 times.
        for try_time in range(1800):
            current_stream = BulkLoadStream(self.client.get_bulkload_stream(self.meta_data.get_schema_name(),
                                                                            self.meta_data.get_table_name(),
                                                                            self.meta_data.get_stream_id()),
                                            self.client)
            state = current_stream.get_stream_state()
            sql_error_msg = current_stream.get_sql_error()
            _logger.info(
                "Get BulkLoadStream:" + self.meta_data.get_stream_id() + ", state:" + state.name + ",time:" + str(
                    try_time))
            if state == BulkLoadState.COMMIT_SUCCESS or state == BulkLoadState.COMMIT_FAILED:
                break
            else:
                time.sleep(2)
        if state != BulkLoadState.COMMIT_SUCCESS:
            _logger.error(
                "BulkLoadStream:" + self.get_stream_id() + " sync commit failed or timeout with state:" + state.name
                + " with error:" + sql_error_msg)
            raise IOError(
                "BulkLoadStream:" + self.get_stream_id() + " sync commit failed or timeout with state:" + state.name
                + " with error:" + sql_error_msg)

    def abort(self):
        _logger.info("Aborting BulkLoadStream:" + self.meta_data.get_stream_id())
        ret = self.client.commit_bulkload_stream(self.meta_data.get_instance_id(), self.meta_data.get_workspace(),
                                                 self.meta_data.get_schema_name(), self.meta_data.get_table_name(),
                                                 self.meta_data.get_stream_id(), '', '',
                                                 BulkLoadCommitMode.ABORT_STREAM)
        if ret.get_state() != BulkLoadState.ABORTED:
            raise IOError(
                "BulkLoadStream:" + self.get_stream_id() + " abort failed ")

    def close(self):
        return



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rm_app_meta.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11rm_app_meta.proto\x12\x11\x63om.clickzetta.rm\"\xaa\x02\n\tRMAppMeta\x12\x13\n\x06\x61pp_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x15\n\x08\x61pp_name\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x15\n\x08priority\x18\x03 \x01(\x05H\x02\x88\x01\x01\x12\x12\n\x05vc_id\x18\x04 \x01(\x03H\x03\x88\x01\x01\x12\x35\n\tapp_state\x18\x05 \x01(\x0e\x32\x1d.com.clickzetta.rm.RMAppStateH\x04\x88\x01\x01\x12\x18\n\x0bsubmit_time\x18\x06 \x01(\x03H\x05\x88\x01\x01\x12\x18\n\x0b\x66inish_time\x18\x07 \x01(\x03H\x06\x88\x01\x01\x42\t\n\x07_app_idB\x0b\n\t_app_nameB\x0b\n\t_priorityB\x08\n\x06_vc_idB\x0c\n\n_app_stateB\x0e\n\x0c_submit_timeB\x0e\n\x0c_finish_time*G\n\nRMAppState\x12\x12\n\x0eRM_APP_RUNNING\x10\x00\x12\x12\n\x0eRM_APP_SUCCESS\x10\x01\x12\x11\n\rRM_APP_FAILED\x10\x02\x42(\n\x17\x63om.clickzetta.rm.protoB\x0bRMAppProtosP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rm_app_meta_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\027com.clickzetta.rm.protoB\013RMAppProtosP\001'
  _RMAPPSTATE._serialized_start=341
  _RMAPPSTATE._serialized_end=412
  _RMAPPMETA._serialized_start=41
  _RMAPPMETA._serialized_end=339
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hash.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nhash.proto\x12\x04kudu*R\n\rHashAlgorithm\x12\x10\n\x0cUNKNOWN_HASH\x10\x00\x12\x11\n\rMURMUR_HASH_2\x10\x01\x12\r\n\tCITY_HASH\x10\x02\x12\r\n\tFAST_HASH\x10\x03\x42\x11\n\x0forg.apache.kudu')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'hash_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017org.apache.kudu'
  _HASHALGORITHM._serialized_start=20
  _HASHALGORITHM._serialized_end=102
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: role_meta.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0frole_meta.proto\x12\x08\x63z.proto\"[\n\x04Role\x12\r\n\x05\x61lias\x18\x01 \x01(\t\x12 \n\x04type\x18\x02 \x01(\x0e\x32\x12.cz.proto.RoleType\x12\"\n\x05level\x18\x03 \x01(\x0e\x32\x13.cz.proto.RoleLevel*+\n\x08RoleType\x12\r\n\tRT_SYSTEM\x10\x00\x12\x10\n\x0cRT_CUSTOMIZE\x10\x01*,\n\tRoleLevel\x12\r\n\tRL_SYSTEM\x10\x00\x12\x10\n\x0cRL_WORKSPACE\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'role_meta_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ROLETYPE._serialized_start=122
  _ROLETYPE._serialized_end=165
  _ROLELEVEL._serialized_start=167
  _ROLELEVEL._serialized_end=211
  _ROLE._serialized_start=29
  _ROLE._serialized_end=120
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: virtual_cluster.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15virtual_cluster.proto\x12\x11\x63om.clickzetta.rm\"\xb3\x02\n\x13\x41nalyticsProperties\x12\x19\n\x0cmin_replicas\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x19\n\x0cmax_replicas\x18\x02 \x01(\x05H\x01\x88\x01\x01\x12(\n\x1bmax_concurrency_per_replica\x18\x03 \x01(\x05H\x02\x88\x01\x01\x12\x39\n\x0cscale_policy\x18\x04 \x01(\x0e\x32\x1e.com.clickzetta.rm.ScalePolicyH\x03\x88\x01\x01\x12\x1b\n\x0epreload_tables\x18\x05 \x01(\tH\x04\x88\x01\x01\x42\x0f\n\r_min_replicasB\x0f\n\r_max_replicasB\x1e\n\x1c_max_concurrency_per_replicaB\x0f\n\r_scale_policyB\x11\n\x0f_preload_tables\"h\n\x11GeneralProperties\x12>\n\x10\x63luster_max_size\x18\x01 \x01(\x0e\x32\x1f.com.clickzetta.rm.VClusterSizeH\x00\x88\x01\x01\x42\x13\n\x11_cluster_max_size\"Z\n\nVCResource\x12\x13\n\x06memory\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12\x1a\n\rvirtual_cores\x18\x02 \x01(\x05H\x01\x88\x01\x01\x42\t\n\x07_memoryB\x10\n\x0e_virtual_cores\"\xd5\x07\n\x18VirtualClusterProperties\x12\x11\n\x04name\x18\x01 \x01(\tH\x01\x88\x01\x01\x12\x18\n\x0binstance_id\x18\x02 \x01(\x03H\x02\x88\x01\x01\x12\x19\n\x0cworkspace_id\x18\x03 \x01(\x03H\x03\x88\x01\x01\x12:\n\x0c\x63luster_type\x18\x04 \x01(\x0e\x32\x1f.com.clickzetta.rm.VClusterTypeH\x04\x88\x01\x01\x12:\n\x0c\x63luster_size\x18\x05 \x01(\x0e\x32\x1f.com.clickzetta.rm.VClusterSizeH\x05\x88\x01\x01\x12\x46\n\x14\x61nalytics_properties\x18\x06 \x01(\x0b\x32&.com.clickzetta.rm.AnalyticsPropertiesH\x00\x12\x42\n\x12general_properties\x18\x07 \x01(\x0b\x32$.com.clickzetta.rm.GeneralPropertiesH\x00\x12\"\n\x15\x61uto_stop_latency_sec\x18\x08 \x01(\x05H\x06\x88\x01\x01\x12\x1f\n\x12\x61uto_start_enabled\x18\t \x01(\x08H\x07\x88\x01\x01\x12\x41\n\x03tag\x18\n \x03(\x0b\x32\x34.com.clickzetta.rm.VirtualClusterProperties.TagEntry\x12\x14\n\x07\x63omment\x18\x0b \x01(\tH\x08\x88\x01\x01\x12)\n\x1cquery_process_time_limit_sec\x18\r \x01(\x05H\t\x88\x01\x01\x12\x1b\n\x0e\x63reate_time_ms\x18\x0e \x01(\x03H\n\x88\x01\x01\x12 \n\x13last_modify_time_ms\x18\x0f \x01(\x03H\x0b\x88\x01\x01\x12\x1c\n\x0f\x63reator_user_id\x18\x10 \x01(\x03H\x0c\x88\x01\x01\x12\x14\n\x07version\x18\x11 \x01(\tH\r\x88\x01\x01\x1a*\n\x08TagEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x10\n\x0eresource_oneofB\x07\n\x05_nameB\x0e\n\x0c_instance_idB\x0f\n\r_workspace_idB\x0f\n\r_cluster_typeB\x0f\n\r_cluster_sizeB\x18\n\x16_auto_stop_latency_secB\x15\n\x13_auto_start_enabledB\n\n\x08_commentB\x1f\n\x1d_query_process_time_limit_secB\x11\n\x0f_create_time_msB\x16\n\x14_last_modify_time_msB\x12\n\x10_creator_user_idB\n\n\x08_version\"e\n\x0bRequestInfo\x12\x17\n\nrequest_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1b\n\x0eoperator_token\x18\x02 \x01(\x0cH\x01\x88\x01\x01\x42\r\n\x0b_request_idB\x11\n\x0f_operator_token\"\xbf\x01\n\x0cResponseInfo\x12\x17\n\nrequest_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12.\n\x06status\x18\x02 \x01(\x0e\x32\x19.com.clickzetta.rm.StatusH\x01\x88\x01\x01\x12\x17\n\nerror_code\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x16\n\terror_msg\x18\x04 \x01(\tH\x03\x88\x01\x01\x42\r\n\x0b_request_idB\t\n\x07_statusB\r\n\x0b_error_codeB\x0c\n\n_error_msg*\xd0\x01\n\x0cVClusterSize\x12\n\n\x06XSMALL\x10\x00\x12\t\n\x05SMALL\x10\x01\x12\n\n\x06MEDIUM\x10\x02\x12\t\n\x05LARGE\x10\x03\x12\n\n\x06XLARGE\x10\x04\x12\x0b\n\x07X2LARGE\x10\x05\x12\x0b\n\x07X3LARGE\x10\x06\x12\x0b\n\x07X4LARGE\x10\x07\x12\x0b\n\x07X5LARGE\x10\x08\x12\x0b\n\x07X6LARGE\x10\t\x12\x0f\n\x0b\x43USTOMIZED3\x10\x65\x12\x10\n\x0c\x43USTOMIZED52\x10\x66\x12\x10\n\x0c\x43USTOMIZED48\x10g\x12\x10\n\x0c\x43USTOMIZED12\x10h**\n\x0cVClusterType\x12\x0b\n\x07GENERAL\x10\x00\x12\r\n\tANALYTICS\x10\x01*\x1b\n\x0bScalePolicy\x12\x0c\n\x08STANDARD\x10\x00*\xb5\x01\n\x13VirtualClusterState\x12\r\n\tSUSPENDED\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\x0c\n\x08STARTING\x10\x02\x12\x0e\n\nSCALING_UP\x10\x03\x12\x10\n\x0cSCALING_DOWN\x10\x04\x12\x0e\n\nSUSPENDING\x10\x05\x12\x0c\n\x08\x44ROPPING\x10\x06\x12\t\n\x05\x45RROR\x10\x07\x12\x0b\n\x07\x44\x45LETED\x10\x08\x12\x0c\n\x08RESUMING\x10\t\x12\x0e\n\nCANCELLING\x10\n*#\n\x06Status\x12\r\n\tSUCCEEDED\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\x42/\n\x17\x63om.clickzetta.rm.protoB\x14VirtualClusterProtosb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'virtual_cluster_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\027com.clickzetta.rm.protoB\024VirtualClusterProtos'
  _VIRTUALCLUSTERPROPERTIES_TAGENTRY._options = None
  _VIRTUALCLUSTERPROPERTIES_TAGENTRY._serialized_options = b'8\001'
  _VCLUSTERSIZE._serialized_start=1834
  _VCLUSTERSIZE._serialized_end=2042
  _VCLUSTERTYPE._serialized_start=2044
  _VCLUSTERTYPE._serialized_end=2086
  _SCALEPOLICY._serialized_start=2088
  _SCALEPOLICY._serialized_end=2115
  _VIRTUALCLUSTERSTATE._serialized_start=2118
  _VIRTUALCLUSTERSTATE._serialized_end=2299
  _STATUS._serialized_start=2301
  _STATUS._serialized_end=2336
  _ANALYTICSPROPERTIES._serialized_start=45
  _ANALYTICSPROPERTIES._serialized_end=352
  _GENERALPROPERTIES._serialized_start=354
  _GENERALPROPERTIES._serialized_end=458
  _VCRESOURCE._serialized_start=460
  _VCRESOURCE._serialized_end=550
  _VIRTUALCLUSTERPROPERTIES._serialized_start=553
  _VIRTUALCLUSTERPROPERTIES._serialized_end=1534
  _VIRTUALCLUSTERPROPERTIES_TAGENTRY._serialized_start=1229
  _VIRTUALCLUSTERPROPERTIES_TAGENTRY._serialized_end=1271
  _REQUESTINFO._serialized_start=1536
  _REQUESTINFO._serialized_end=1637
  _RESPONSEINFO._serialized_start=1640
  _RESPONSEINFO._serialized_end=1831
# @@protoc_insertion_point(module_scope)



import datetime
import decimal
import string
import uuid

import pyarrow as pa
import pyarrow.parquet as pq
import pyarrow.compute as pc
from logging import getLogger

from clickzetta.proto.generated import ingestion_pb2, data_type_pb2, file_format_type_pb2
from clickzetta.client import Client
from clickzetta.bulkload.bulkload_enums import BulkLoadMetaData, BulkLoadConfig, BulkLoadState, FileFormatType

_logger = getLogger(__name__)


class Row:
    def __init__(self, full_fields: dict, table_name: string):
        self.field_name_values = {}
        self.full_fields = full_fields
        self.table_name = table_name

    def set_value(self, field_name: string, field_value):
        if field_name in self.full_fields:
            self.field_name_values[field_name] = field_value
        else:
            raise RuntimeError('Field name:{} is not in table:{}'.format(field_name, self.table_name))


class BulkLoadWriter:
    def __init__(self, client: Client, meta_data: BulkLoadMetaData, config: BulkLoadConfig, partition_id: int):
        self.client = client
        self.meta_data = meta_data
        self.config = config
        self.partition_id = partition_id
        self.file_io = config.get_staging_config().create_file_io()
        self.file_format = config.get_file_format()
        self.location = config.get_staging_config().path
        self.max_file_records = config.get_max_rows_per_file()
        self.max_file_size = config.get_max_file_size_per_file()
        if meta_data.get_state() != BulkLoadState.CREATED:
            raise AssertionError("Failed to create BulkLoadStream due to invalid state:" + meta_data.get_state())
        self.partition_spec = self._parse_partition_spec(self.meta_data.table.schema,
                                                         self.meta_data.get_partition_specs())
        self.finished_files = []
        self.finished_file_sizes = []
        self.file_name_uuid = uuid.uuid4()
        self.file_id = 0
        self.closed = False
        self.current_total_rows = 0
        self.current_total_size = 0
        self.current_record_batch = {}
        self.current_batch_rows = 0
        self.pyarrow_schema = self._generate_pyarrow_schema(self.meta_data.get_table().schema)
        self.writer = None

    def get_stream_id(self):
        return self.meta_data.get_stream_id()

    def get_operation(self):
        return self.meta_data.get_operation()

    def get_schema(self):
        return self.meta_data.get_table().schema

    def get_table(self):
        return self.meta_data.get_table()

    def get_partition_id(self):
        return self.partition_id

    def create_row(self):
        return Row(self.meta_data.get_table().schema, self.meta_data.get_table_name())

    def write(self, row: Row):
        self._check_file_status()
        if not self.current_record_batch:
            self._construct_new_record_batch()

        for partition in self.partition_spec:
            row.field_name_values[partition] = self.partition_spec[partition]

        for filed_name in row.field_name_values:
            self.current_record_batch[filed_name].append(row.field_name_values[filed_name])
        self.current_batch_rows = self.current_batch_rows + 1
        self.current_total_rows = self.current_total_rows + 1

        # batch rows set to 2000, will get from session config later
        if self.current_batch_rows == 2000:
            buffer_size = self._flush_record_batch()
            self.current_total_size = self.current_total_size + buffer_size

    def finish(self):
        if self.closed:
            raise AssertionError('BulkLoadWriter is already closed.')
        self._close_current_file()
        status = self.client.finish_bulkload_stream_writer(self.meta_data.get_instance_id(),
                                                           self.meta_data.get_workspace(),
                                                           self.meta_data.get_schema_name(),
                                                           self.meta_data.get_table_name(),
                                                           self.meta_data.get_stream_id(), self.partition_id,
                                                           self.finished_files, self.finished_file_sizes)
        if status.code != ingestion_pb2.Code.SUCCESS:
            msg = status.message
            raise RuntimeError(
                'Finish BulkLoadStreamWriter for instance:{}, workspace:{}, '
                'table:{}.{}, streamId:{}, partitionId:{} failed. Error message:{}'.format(
                    self.meta_data.get_instance_id(),
                    self.meta_data.get_workspace(), self.meta_data.get_schema_name(), self.meta_data.get_table_name(),
                    self.meta_data.get_stream_id(), self.partition_id, msg))
        _logger.info("Flush bulk load stream {} partitionId {} with {} files".format(self.meta_data.get_stream_id(),
                                                                                     self.partition_id,
                                                                                     len(self.finished_files)))
        self.finished_files.clear()
        self.finished_file_sizes.clear()
        self.closed = True

    def abort(self):
        self.writer.close()
        self.current_total_rows = 0
        self.current_total_size = 0
        self.current_batch_rows = 0
        self.current_record_batch = {}
        self.file_io.close()
        self.closed = True

    def close(self):
        if self.closed:
            return
        self.finish()

    def _flush_record_batch(self):
        batch_data = []
        for item in self.current_record_batch:
            batch_data.append(self._convert_data_type(item, self.current_record_batch[item]))
        batch = pa.record_batch(batch_data, schema=self.pyarrow_schema)
        self.writer.write_batch(batch)
        self.current_batch_rows = 0
        self.current_record_batch.clear()
        return batch.get_total_buffer_size()

    def _convert_data_type(self, field_name: string, field_data: list):
        schema = self.meta_data.get_table().schema
        data_type = schema[field_name]
        if data_type.category == data_type_pb2.DataTypeCategory.INT8:
            return pc.cast(field_data, pa.int8())
        elif data_type.category == data_type_pb2.DataTypeCategory.INT16:
            return pc.cast(field_data, pa.int16())
        elif data_type.category == data_type_pb2.DataTypeCategory.INT32:
            return pc.cast(field_data, pa.int32())
        elif data_type.category == data_type_pb2.DataTypeCategory.INT64:
            return pc.cast(field_data, pa.int64())
        elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT32:
            return pc.cast(field_data, pa.float32())
        elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT64:
            return pc.cast(field_data, pa.float64())
        elif data_type.category == data_type_pb2.DataTypeCategory.DECIMAL:
            precision = data_type.decimalTypeInfo.precision
            scale = data_type.decimalTypeInfo.scale
            return pc.cast(field_data, pa.decimal128(precision, scale))
        elif data_type.category == data_type_pb2.DataTypeCategory.BOOLEAN:
            return pc.cast(field_data, pa.bool_())
        elif data_type.category == data_type_pb2.DataTypeCategory.CHAR or \
                data_type.category == data_type_pb2.DataTypeCategory.VARCHAR or \
                data_type.category == data_type_pb2.DataTypeCategory.STRING:
            return pc.cast(field_data, pa.string())
        elif data_type.category == data_type_pb2.DataTypeCategory.DATE:
            return pc.cast(field_data, pa.date32())
        elif data_type.category == data_type_pb2.DataTypeCategory.TIMESTAMP_LTZ:
            timestamp_unit = data_type.timestamp_info.tsUnit
            if timestamp_unit == data_type_pb2.TimestampUnit.SECONDS:
                return pc.cast(field_data, pa.timestamp('s', tz='UTC'))
            elif timestamp_unit == data_type_pb2.TimestampUnit.MILLISECONDS:
                return pc.cast(field_data, pa.timestamp('ms', tz='UTC'))
            elif timestamp_unit == data_type_pb2.TimestampUnit.MICROSECONDS:
                return pc.cast(field_data, pa.timestamp('us', tz='UTC'))
            elif timestamp_unit == data_type_pb2.TimestampUnit.NANOSECONDS:
                return pc.cast(field_data, pa.timestamp('ns', tz='UTC'))

    def _construct_new_record_batch(self):
        self.current_record_batch.clear()
        table_fields = self.meta_data.get_table().schema
        for filed in table_fields:
            self.current_record_batch[filed] = []
        self.current_batch_rows = 0

    def _generate_pyarrow_schema(self, schema: dict) -> pa.Schema:
        pyarrow_fields = []
        for field in schema:
            data_type = schema[field]
            if data_type.category == data_type_pb2.DataTypeCategory.INT8:
                pyarrow_fields.append(pa.field(field, pa.int8()))
            elif data_type.category == data_type_pb2.DataTypeCategory.INT16:
                pyarrow_fields.append(pa.field(field, pa.int16()))
            elif data_type.category == data_type_pb2.DataTypeCategory.INT32:
                pyarrow_fields.append(pa.field(field, pa.int32()))
            elif data_type.category == data_type_pb2.DataTypeCategory.INT64:
                pyarrow_fields.append(pa.field(field, pa.int64()))
            elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT32:
                pyarrow_fields.append(pa.field(field, pa.float32()))
            elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT64:
                pyarrow_fields.append(pa.field(field, pa.float64()))
            elif data_type.category == data_type_pb2.DataTypeCategory.DECIMAL:
                precision = data_type.decimalTypeInfo.precision
                scale = data_type.decimalTypeInfo.scale
                pyarrow_fields.append(pa.field(field, pa.decimal128(precision, scale)))
            elif data_type.category == data_type_pb2.DataTypeCategory.BOOLEAN:
                pyarrow_fields.append(pa.field(field, pa.bool_()))
            elif data_type.category == data_type_pb2.DataTypeCategory.CHAR or \
                    data_type.category == data_type_pb2.DataTypeCategory.VARCHAR or \
                    data_type.category == data_type_pb2.DataTypeCategory.STRING:
                pyarrow_fields.append(pa.field(field, pa.string()))
            elif data_type.category == data_type_pb2.DataTypeCategory.DATE:
                pyarrow_fields.append(pa.field(field, pa.date32()))
            elif data_type.category == data_type_pb2.DataTypeCategory.TIMESTAMP_LTZ:
                timestamp_unit = data_type.timestamp_info.tsUnit
                if timestamp_unit == data_type_pb2.TimestampUnit.SECONDS:
                    pyarrow_fields.append(pa.field(field, pa.timestamp('s', tz='UTC')))
                elif timestamp_unit == data_type_pb2.TimestampUnit.MILLISECONDS:
                    pyarrow_fields.append(pa.field(field, pa.timestamp('ms', tz='UTC')))
                elif timestamp_unit == data_type_pb2.TimestampUnit.MICROSECONDS:
                    pyarrow_fields.append(pa.field(field, pa.timestamp('us', tz='UTC')))
                elif timestamp_unit == data_type_pb2.TimestampUnit.NANOSECONDS:
                    pyarrow_fields.append(pa.field(field, pa.timestamp('ns', tz='UTC')))
        return pa.schema(pyarrow_fields)

    def _parse_partition_spec(self, schema: dict, partition_spec: string) -> dict:
        partition_value_dict = {}
        if partition_spec.strip() == '':
            return partition_value_dict
        partition_pairs = partition_spec.strip().split(',')
        for partition in partition_pairs:
            kv = partition.strip().split('=')
            partition_value_dict[kv[0].strip()] = kv[1].strip()
        for field in schema:
            if field in partition_value_dict:
                data_type = schema[field]
                if data_type.category == data_type_pb2.DataTypeCategory.INT8 or \
                        data_type.category == data_type_pb2.DataTypeCategory.INT16 or \
                        data_type.category == data_type_pb2.DataTypeCategory.INT32:
                    partition_value_dict[field] = int(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.INT64:
                    partition_value_dict[field] = int(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT32 or \
                        data_type.category == data_type_pb2.DataTypeCategory.FLOAT64:
                    partition_value_dict[field] = float(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.DECIMAL:
                    partition_value_dict[field] = decimal.Decimal(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.BOOLEAN:
                    partition_value_dict[field] = bool(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.DATE:
                    partition_value_dict[field] = (
                            datetime.strptime(partition_value_dict[field], '%Y-%m-%d') - datetime(1970, 1, 1)).days

        return partition_value_dict

    def _get_current_file_name(self):
        return '{}{}-{}.{}'.format(self.location, self.file_name_uuid, self.file_id, self.file_format.value)

    def _create_next_file_writer(self):
        file_name = self._get_current_file_name()
        if self.file_format == FileFormatType.PARQUET:
            return pq.ParquetWriter(file_name, self.pyarrow_schema, filesystem=self.file_io)
        else:
            raise NotImplementedError('File format:{} is not supported yet.'.format(self.file_format.value))

    def _close_current_file(self):
        if self.writer is not None:
            buffer_size = self._flush_record_batch()
            self.writer.close()
            file_name = self._get_current_file_name()
            self.finished_files.append(file_name)
            self.finished_file_sizes.append(self.current_total_size + buffer_size)
            self.current_total_size = 0
            self.current_total_rows = 0
            self.current_batch_rows = 0
            self.current_record_batch.clear()
            self.file_id = self.file_id + 1
            self.writer = None

    def _check_file_status(self):
        if self.writer is not None:
            if self.current_total_rows >= self.max_file_records or self.current_total_size > self.max_file_size:
                self._close_current_file()

        if self.writer is None:
            self.writer = self._create_next_file_writer()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: virtual_cluster_meta.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import virtual_cluster_pb2 as virtual__cluster__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1avirtual_cluster_meta.proto\x12\x08\x63z.proto\x1a\x15virtual_cluster.proto\"\xdf\x02\n\x13\x41nalyticsProperties\x12\x19\n\x0cmin_replicas\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x19\n\x0cmax_replicas\x18\x02 \x01(\x05H\x01\x88\x01\x01\x12(\n\x1bmax_concurrency_per_replica\x18\x03 \x01(\x05H\x02\x88\x01\x01\x12\x39\n\x0cscale_policy\x18\x04 \x01(\x0e\x32\x1e.com.clickzetta.rm.ScalePolicyH\x03\x88\x01\x01\x12\x19\n\x0c\x63ur_replicas\x18\x05 \x01(\x05H\x04\x88\x01\x01\x12\x1b\n\x0epreload_tables\x18\x06 \x01(\tH\x05\x88\x01\x01\x42\x0f\n\r_min_replicasB\x0f\n\r_max_replicasB\x1e\n\x1c_max_concurrency_per_replicaB\x0f\n\r_scale_policyB\x0f\n\r_cur_replicasB\x11\n\x0f_preload_tables\"h\n\x11GeneralProperties\x12>\n\x10\x63luster_max_size\x18\x01 \x01(\x0e\x32\x1f.com.clickzetta.rm.VClusterSizeH\x00\x88\x01\x01\x42\x13\n\x11_cluster_max_size\"c\n\x07JobInfo\x12\x19\n\x0cjobs_running\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x1a\n\rjobs_in_queue\x18\x02 \x01(\x05H\x01\x88\x01\x01\x42\x0f\n\r_jobs_runningB\x10\n\x0e_jobs_in_queue\"\xcb\x07\n\x12VirtualClusterMeta\x12\x35\n\x0c\x63luster_type\x18\x01 \x01(\x0e\x32\x1f.com.clickzetta.rm.VClusterType\x12:\n\x0c\x63luster_size\x18\x02 \x01(\x0e\x32\x1f.com.clickzetta.rm.VClusterSizeH\x01\x88\x01\x01\x12=\n\x14\x61nalytics_properties\x18\x03 \x01(\x0b\x32\x1d.cz.proto.AnalyticsPropertiesH\x00\x12\x39\n\x12general_properties\x18\x04 \x01(\x0b\x32\x1b.cz.proto.GeneralPropertiesH\x00\x12\"\n\x15\x61uto_stop_latency_sec\x18\x05 \x01(\x05H\x02\x88\x01\x01\x12\x1f\n\x12\x61uto_start_enabled\x18\x06 \x01(\x08H\x03\x88\x01\x01\x12\x32\n\x03tag\x18\x07 \x03(\x0b\x32%.cz.proto.VirtualClusterMeta.TagEntry\x12)\n\x1cquery_process_time_limit_sec\x18\t \x01(\x05H\x04\x88\x01\x01\x12:\n\x05state\x18\n \x01(\x0e\x32&.com.clickzetta.rm.VirtualClusterStateH\x05\x88\x01\x01\x12>\n\tpre_state\x18\x0b \x01(\x0e\x32&.com.clickzetta.rm.VirtualClusterStateH\x06\x88\x01\x01\x12\x16\n\terror_msg\x18\x0c \x01(\tH\x07\x88\x01\x01\x12(\n\x08job_info\x18\r \x01(\x0b\x32\x11.cz.proto.JobInfoH\x08\x88\x01\x01\x12\x19\n\x0cworkspace_id\x18\x0e \x01(\x03H\t\x88\x01\x01\x12\x12\n\x05vc_id\x18\x0f \x01(\x03H\n\x88\x01\x01\x12\x17\n\nstate_info\x18\x10 \x01(\tH\x0b\x88\x01\x01\x12\x14\n\x07version\x18\x11 \x01(\tH\x0c\x88\x01\x01\x1a*\n\x08TagEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x10\n\x0eresource_oneofB\x0f\n\r_cluster_sizeB\x18\n\x16_auto_stop_latency_secB\x15\n\x13_auto_start_enabledB\x1f\n\x1d_query_process_time_limit_secB\x08\n\x06_stateB\x0c\n\n_pre_stateB\x0c\n\n_error_msgB\x0b\n\t_job_infoB\x0f\n\r_workspace_idB\x08\n\x06_vc_idB\r\n\x0b_state_infoB\n\n\x08_versionb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'virtual_cluster_meta_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _VIRTUALCLUSTERMETA_TAGENTRY._options = None
  _VIRTUALCLUSTERMETA_TAGENTRY._serialized_options = b'8\001'
  _ANALYTICSPROPERTIES._serialized_start=64
  _ANALYTICSPROPERTIES._serialized_end=415
  _GENERALPROPERTIES._serialized_start=417
  _GENERALPROPERTIES._serialized_end=521
  _JOBINFO._serialized_start=523
  _JOBINFO._serialized_end=622
  _VIRTUALCLUSTERMETA._serialized_start=625
  _VIRTUALCLUSTERMETA._serialized_end=1596
  _VIRTUALCLUSTERMETA_TAGENTRY._serialized_start=1332
  _VIRTUALCLUSTERMETA_TAGENTRY._serialized_end=1374
# @@protoc_insertion_point(module_scope)
